# -*- coding: utf-8 -*-
from odoo import api, fields, models, _
from odoo.tools import format_date
from odoo.exceptions import UserError
from dateutil.relativedelta import relativedelta


class SaleOrderLine(models.Model):
    _inherit = 'sale.order.line'

    # Prevent altering price unit of recurring product
    is_recurring_not_editable = fields.Boolean(compute="_compute_is_recurring_not_editable")

    @api.depends('recurring_invoice', 'product_id')
    def _compute_is_recurring_not_editable(self):

        # Products Exceptions
        exceptions_product_ids = [
            self.env.ref('openerp_enterprise.product_light_user_discount').id,
            self.env.ref('openerp_enterprise.product_multiyear_discount').id,
            self.env.ref('openerp_enterprise.product_odoo_subscription_extra_discount').id,
            self.env.ref('openerp_enterprise.product_learning_partner_discount').id,
            self.env.ref('openerp_enterprise.product_official_partner_discount').id,
            self.env.ref('openerp_enterprise.product_maintenance_adjustment').id
        ]

        for line in self:
            line.is_recurring_not_editable = line.recurring_invoice and line.product_id.id not in exceptions_product_ids


    @api.depends('product_id', 'order_id.sale_order_template_id')
    def _compute_product_uom(self):
        # product categories that should have its uom updated
        product_category_ids = [
            self.env.ref('openerp_enterprise.product_category_odoo_users'),  # Odoo Users
            self.env.ref('openerp_enterprise.category_openerp_apps'),  # Odoo Apps
            self.env.ref('openerp_enterprise.product_category_odoo_sh'), # Odoo.sh
            self.env.ref('openerp_enterprise.category_maintenance'),
        ]

        for l in self:
            sub_template = l.order_id.sale_order_template_id
            if sub_template and l.product_id.categ_id in product_category_ids:
                l.product_uom = sub_template.get_product_uom(l.product_id)
            else:
                super(SaleOrderLine, l)._compute_product_uom()

    @api.onchange('product_id')
    def product_id_change(self):
        """
        Try to detect the most appropriate uom when using subscription products.

        To avoid a complete UoM mess with multi-years subscriptions, this
        override fetch the uom from the subscription template or from a linked subscription
        instead of the default one of the product. This avoids tedious manipulations of uoms
        that salespeople will probably mess up anyway.
        """
        if self.product_id and self.order_id.sale_order_template_id:
            self.product_uom = self.order_id.sale_order_template_id.get_product_uom(self.product_id)

    def _timesheet_create_task(self, project):
        task = super(SaleOrderLine, self)._timesheet_create_task(project)
        if project.privacy_visibility == 'portal':
            # auto-subscribe customer to the task when project visibility is set
            # to portal, so that they can see the task from /my/tasks
            task.message_subscribe(partner_ids=[self.order_partner_id.id])
        return task

    def _prepare_subscription_line_data(self):
        partner_sub_discount_product_ids = [
            self.env.ref('openerp_enterprise.product_learning_partner_discount').id,
            self.env.ref('openerp_enterprise.product_official_partner_discount').id,
        ]
        res = super(SaleOrderLine, self)._prepare_subscription_line_data()
        Product = self.env['product.product']
        UoM = self.env["uom.uom"]
        pricelist = self[0].order_id.pricelist_id
        for line in res:
            if line[2]['product_id'] in partner_sub_discount_product_ids:
                pdt = Product.browse(line[2]['product_id'])
                uom = UoM.browse(line[2]['uom_id'])
                price_unit = pricelist._get_product_price(pdt, 1, uom=uom)
                line[2]['price_unit'] = price_unit
        return res

    def _get_protected_fields(self):
        # allow `set_apps_recurring_lines` to update a subscription when needed
        if self.env.su:
            return []
        # otherwise let the code decide
        return super()._get_protected_fields()

    def _is_not_sellable_line(self):
        user_discounts = [
            self.env.ref('openerp_enterprise.product_odoo_user_discount').id,
            self.env.ref('openerp_enterprise.product_odoo_user_standard_discount').id,
            self.env.ref('openerp_enterprise.product_odoo_user_custom_discount').id,
        ]
        return self.product_id.id in user_discounts or super()._is_not_sellable_line()

    def _reduce_discount_7pc(self):
        self.ensure_one()
        pc7 = -self.order_id.amount_untaxed * .07
        if pc7 >= self.price_unit:
            self.order_id.message_post(
                body="Pricing Compensation reduced of %.2f (7%% of %.2f)" % (pc7, self.order_id.amount_untaxed)
            )
            self.price_unit -= pc7
            return pc7
        else:
            old_pu = self.price_unit
            self.order_id.message_post(
                body="Pricing Compensation removed. (%.2f < %.2f - 7%% of %.2f)" % (
                    pc7, self.price_unit, self.order_id.amount_untaxed
                )
            )
            self.unlink()
            return old_pu

    def _prepare_invoice_line(self, **optional_values):
        self.ensure_one()
        res = super()._prepare_invoice_line(**optional_values)
        # for event ticket/booth force revenues recognitation to the same
        # period as the related event.
        if self.product_type in ('event', 'event_booth') and self.event_id:
            event = self.event_id._set_tz_context()
            begin_tz = fields.Datetime.context_timestamp(event, event.date_begin)
            end_tz = fields.Datetime.context_timestamp(event, event.date_end)
            res.update(
                deferred_start_date=fields.Date.to_string(begin_tz),
                deferred_end_date=fields.Date.to_string(end_tz),
            )
        elif self.order_id.is_subscription or self.order_id.subscription_state == '7_upsell':
            # for first year discount on yearly subscription do actually
            # only recognize the revenue on the first year.
            sub = self.order_id
            product_is_user_discount = self.product_id in (
                self.env.ref('openerp_enterprise.product_odoo_user_discount'),
                self.env.ref('openerp_enterprise.product_odoo_user_standard_discount'),
                self.env.ref('openerp_enterprise.product_odoo_user_custom_discount'),
            )
            if sub.plan_id.billing_period_unit == 'year' and sub.first_contract_date and product_is_user_discount:
                deferred_start_date = sub.start_date  # upsell date or start/renewal date
                deferred_end_date = sub.first_contract_date + relativedelta(years=1, days=-1)
                if deferred_end_date < deferred_start_date:
                    # we're trying to invoice for a discount that shouldn't be there anymore
                    raise UserError(_("Cannot invoice Initial User discount that have expired"))

                # also update duration & dates in description
                lang_code = self.order_id.partner_id.lang
                env_partner_lang = self.env["base"].with_context(lang=lang_code).env
                duration_labels = dict(sub.plan_id._fields['billing_period_unit']._description_selection(env_partner_lang))
                duration = f"1 {duration_labels['year']}"
                format_start = format_date(self.env, deferred_start_date, lang_code=lang_code)
                format_next = format_date(self.env, deferred_end_date, lang_code=lang_code)
                start_to_next = _("\n%s to %s", format_start, format_next)
                description = f"{self.name} - {duration}{start_to_next}"
                res.update(
                    deferred_start_date=deferred_start_date,
                    deferred_end_date=deferred_end_date,
                    name=description,
                )
        return res
